/*******************************************/
/* To compile type: 'make'                 */
/* To run type: ./APYCore genofile cores   */
/*******************************************/
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <unordered_map>
#include <omp.h>
#include <mkl.h>
#include "mkl_lapacke.h"

#include "pedigree.h"

using namespace std;

/*****************************************************************/
/*****                Start of Program                       *****/
/*****************************************************************/
int main(int argc, char* argv[])
{
    time_t full_begin_time = time(0);
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+--" << endl;
    cout <<"|                                    ///\\\\\\        |" << endl;
    cout <<"|                                   /      \\       |" << endl;
    cout <<"|  APY Core Count Program          #-/.\\==/.\\      |" << endl;
    cout <<"|  Jeremy T. Howard               (, \\_/ \\\\_/      |" << endl;
    cout <<"|  Date: Oct. 20, 2018             |    -`  |      |" << endl;
    cout <<"|                                   \\  \\_/ /       |" << endl;
    cout <<"|                                  /`-.__.`        |" << endl;
    cout <<"|                               .-'`-.___|__       |" << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+--" << endl;
    if(argc != 3){cout << "Incorrect Number of parameters specified" << endl; exit (EXIT_FAILURE);}
    /* Generate Output File Name Strings based on Input Parameters */
    string genotype_file = argv[1];
    int cores = atoi(argv[2]);
    /* Number of threads */
    omp_set_num_threads(cores);
    mkl_set_num_threads_local(cores);
    cout << "- Parameters Specified:"<<endl;
    cout << "   - Genotype file name is '" << genotype_file << "'." << endl;
    cout << "       - Assume that: " << endl;
    cout << "           - Column 1 is Animal; Column 2 is Genotype." << endl;
    cout << "           - No header in file and any space delimeter." << endl << endl;
    /****************************************************/
    /* Read in genotype file and save to pedigree class */
    /****************************************************/
    vector < Pedigree_Class > pedigree;                 /* Save information about animal in vector of "Pedigree_Class" */
    unordered_map <string,int> linker;                  /* Hash-table data structure for faster lookup */
    std:unordered_map<string,int>::const_iterator got;  /* Search if key exists */
    string line,tempanimal;                             /* Initialize temp variables */
    /** Expect all genotypes string to be the same length */
    int numbergeno; int numberanimals = 0; int linenumber = 0;
    cout << "- Read in genotype file: " << endl;
    ifstream infile1;
    infile1.open(genotype_file.c_str());
    if(infile1.fail()){cout << "Genotype File Doesn't Exist!\n"; exit (EXIT_FAILURE);}
    while (getline(infile1,line))
    {
        size_t pos = line.find("\t", 0);                 // Find position where delimiter is at
        if(pos != std::string::npos){cout << endl << "Delimeter in genotype file is a tab!!" << endl; exit (EXIT_FAILURE);}
        else{
            /* First split it apart */
            size_t pos = line.find(" ",0); tempanimal = line.substr(0,pos); line.erase(0, pos + 1);
            int skipcount = 0;
            while(1)
            {
                size_t pos = line.find(" ",0);
                if(line.substr(0,pos) != "")
                {
                    got = linker.find(tempanimal);
                    if(got == linker.end()){
                        /* Remove hanging spacers if applicable */
                        line.erase(remove(line.begin(), line.end(), ' '),line.end());
                        Pedigree_Class temppedigree;
                        temppedigree.upd_animal(tempanimal);
                        temppedigree.upd_genotype(line);
                        pedigree.push_back(temppedigree); numberanimals++;
                        linker.insert({tempanimal,pedigree.size()-1});
                        /* First animal sets the snp count length */
                        if(linenumber == 0){
                            numbergeno = (pedigree[0].get_genotype()).size();
                        } else {
                            if((pedigree[pedigree.size()-1].get_genotype()).size() != numbergeno)
                            {
                                cout << "\nNumber of SNP are not the same across animals. Check line number ";
                                cout << linenumber << "." << endl; exit (EXIT_FAILURE);
                            }
                        }
                        break;
                    } else {
                        cout << "\nAnimal appears more than once in genotype file: '" << tempanimal << "'" << endl;
                        exit (EXIT_FAILURE);
                    }
                }
                if(line.substr(0,pos) != "" && skipcount == 0){line.erase(0, pos); skipcount++;}
                if(line.substr(0,pos) == ""){line.erase(0, pos + 1);}
            }
        }
        linenumber++;
    }
    cout << "   - Number of genotyped animals read in: " << numberanimals << endl;
    cout << "   - Size of genotype string: " << numbergeno << endl << endl;
    /*****************************************************/
    /** Begin to calculate Singular values using dgesvd **/
    /*****************************************************/
    time_t sub_begin_time = time(0);
    cout << "- Begin Calculating Singular Values (SV; SV^2 = eigen values)." << endl;
    cout << "   - Begin Constructing Z Matrix: " << endl;
    cout << "   - Calculate Allele Frequencies." << endl;
    vector < double > meangeno(numbergeno,0.0);    /* allele frequency */
    vector < int > numgeno(numbergeno,0);          /* number of animals with non-missing genotypes */
    vector < vector < double >> M(3, vector<double>(numbergeno,0.0)); /* M matrix */
    string tempgeno;    /* used as a temporary variable for genotype string */
    for(int i = 0; i < numberanimals; i++)
    {
        tempgeno = pedigree[i].get_genotype();
        for(int j = 0; j < numbergeno; j++)
        {
            int tempa = tempgeno[j] - 48;                   /* ASCI value is 48 for 0; and ASCI value for 0 to 9 is 48 to */
            if(tempa == 3 || tempa == 4){ tempa = 1;}       /* Convert 3 or 4 to 1 */
            if(tempa != 5){meangeno[j] = meangeno[j] + tempa; numgeno[j] = numgeno[j] + 1;}
        }
    }
    cout << "   - Generate M Matrix." << endl;
    for(int i = 0; i < meangeno.size(); i++){meangeno[i] = (meangeno[i] / (2 * double(numgeno[i])));}
    //for(int i = 0; i < 20; i++){cout << meangeno[i] << endl;}
    for(int i = 0; i < 3; i++)
    {
        for(int j = 0; j < meangeno.size(); j++){M[i][j] = i - (2 * meangeno[j]);}
    }
    //for(int i = 0; i < 20; i++)
    //{
    //    for(int j = 0; j < 3; j++){cout << M[j][i] << " ";}
    //    cout << endl;
    //}
    cout << "   - Generate Z Matrix (Multi-threaded)." << endl;
    double *_Z=new double[numbergeno*numberanimals]; // alloc memory to Z
    int tempj = 0; int tempi = 0;
    #pragma omp parallel for private(tempj)
    for(tempi = 0; tempi < numberanimals; tempi++)
    {
        string tempgenoa = pedigree[tempi].get_genotype();
        for(tempj =0; tempj < numbergeno; tempj++)
        {
            int tempa = tempgenoa[tempj] - 48;
            if(tempa == 3 || tempa == 4){tempa = 1;}
            if(tempa != 5){_Z[(tempj*numberanimals)+tempi] = M[tempa][tempj];}
            if(tempa == 5){_Z[(tempj*numberanimals)+tempi] = 0;}
        }
    }
    //for(int i = 0; i < 20; i++)
    //{
    //    for(int j = 0; j < 10; j++){cout << _Z[(i*numbergeno)+j] << " ";}
    //    cout << endl;
    //}
    cout << "   - Begin using Intel 'dgesvd' to compute singular values (Number of Threads: ";
    cout << cores << ")." << endl;
    double *s = new double[numberanimals];                          // --- Singular values
    double *u = new double[numbergeno*numbergeno];                  // --- Matrices of LEFT singular vectors
    double *vt = new double[numberanimals*numberanimals];           // --- Matrices of RIGHT singular vectors
    long long int m = int(numbergeno);
    long long int n = int(numberanimals);
    long long int info = int(0);
    double superb[numberanimals-1];
    LAPACKE_dgesvd(LAPACK_ROW_MAJOR,'N','N',m,n,_Z,n,s,u,m,vt,n,superb);
    if(info > 0)
    {
        cout << "The algorithm computing SVD failed to converge.\n";
        exit (EXIT_FAILURE);
    }
    cout << "- Finished calculating singular values.";
    delete[] _Z; delete[] u; delete[] vt;
    time_t sub_end_time = time(0);
    cout << " (Took: "<<difftime(sub_end_time,sub_begin_time)<<" seconds)"<<endl<<endl;
    /*******************************************************************/
    /** Now Find # eigenvalues explain 90, 95, 98 and 99 % and output **/
    /*******************************************************************/
    cout << "- Calculate Summary Statistics and Output Results." << endl;
    vector < double > eigenztz(numberanimals,0.0);
    vector < double > cumulativesum(numberanimals,0.0);
    vector < double > percexplained(numberanimals,0.0);
    /* singluar values^2 = eigenvalues */
    double cumsum = 0.0;
    for(int i = 0; i < numberanimals; i++)
    {
        eigenztz[i] = s[i]*s[i];
        cumsum += eigenztz[i];
        cumulativesum[i] = cumsum;
    }
    double sum = 0.0;
    for(int i = 0; i < numberanimals; i++){sum += eigenztz[i];}
    for(int i = 0; i < numberanimals; i++){percexplained[i] = cumulativesum[i] / double(sum);}
    /*****************************************/
    /* First output all svd and eigen values */
    /*****************************************/
    /* Remove file if it available */
    fstream checkOut; checkOut.open("snp_svd", std::fstream::out | std::fstream::trunc); checkOut.close();
    std::ofstream output("snp_svd", std::ios_base::app | std::ios_base::out);
    output << "NumberExcluded singular_val eigZtZ cumsum PercExplained" << endl;
    for(int i = 0; i < numberanimals; i++)
    {
        output << i + 1 << " " << s[i] << " " << eigenztz[i] << " " << cumulativesum[i] << " ";
        output << percexplained[i] << endl;
    }
    string found90 = "no"; string found95 = "no"; string found98 = "no"; string found99 = "no";
    for(int i = 0; i < numberanimals; i++)
    {
        if(percexplained[i] > 0.90 && found90 == "no"){cout << "   - EIG 90% " << i + 1 << endl; found90 = "yes";}
        if(percexplained[i] > 0.95 && found95 == "no"){cout << "   - EIG 95% " << i + 1 << endl; found95 = "yes";}
        if(percexplained[i] > 0.98 && found98 == "no"){cout << "   - EIG 98% " << i + 1 << endl; found98 = "yes";}
        if(percexplained[i] > 0.99 && found99 == "no"){cout << "   - EIG 99% " << i + 1 << endl; found99 = "yes";}
    }
    cout << endl;
    delete[] s; mkl_free_buffers(); mkl_thread_free_buffers();
    time_t full_end_time = time(0);
    cout << "- Finished (Took: "<<difftime(full_end_time,full_begin_time)<<" seconds)"<<endl<<endl;
}
